################ Classes ################
print("------------------ Classes ------------------")
class CreditCard(object):
    name = "Chase Credit card" # <- this is a class variable

    # defines how to create an instance of the class
    def __init__(self, yearly_rate):
        self.balance = 0
        self.yearly_rate = yearly_rate
        self.transactions = []
    
    ## getter method 
    def get_balance(self):
        return self.balance
    
    def get_yearly_rate(self):
        return self.yearly_rate
    
    # setter method
    def set_yearly_rate(self, yearly_rate):
        self.yearly_rate = yearly_rate
    
    def spend(self, amount):
        self.transactions.append(amount)
        self.balance += amount
        
    def pay_bill(self, amount):
        self.balance -= amount
    
    def compare(self, other):
        return self < other
        
    def monthly_update(self):
        interests = self.get_balance()*self.yearly_rate/12.0/100.0
        self.balance += interests

    def __lt__(self, other):
        # called when we call the < opperator between two instances of Credit card
        return self.balance <= other.balance
        
    def __str__(self):
        # called when we call the print on an instance of CreditCard
        return "amount due: "+str(self.get_balance())

## What is the data attributes for the CreditCard class? yearly_rate
## What is the class variable for the CreditCard class? name
## We can access a class variable without creating an instance of the class as follows
print(CreditCard.name)

my_card = CreditCard(0.10) # creates an instance of Credit card 
my_other_card = CreditCard(0.15) # creates an instance of Credit card 
# here CreditCard is the class, my_card and my other card are SPECIFIC instances of that class
# similar to how int is the class, 5 is an instance of int

# we can perform the procedures defined in the class on both instances of Credit card
# note here that because we are calling the functions on the instance of that class, 
# we dont need to provide an equivalent of the self parameter. 
my_card.spend(1.99) # calls the "spend" method on my_card
my_card.spend(3.99) # calls the "spend" method on my_card
my_card.monthly_update() # calls the "monthly_update" method on my_card
my_card.pay_bill(4) # calls the "pay_bill" method on my_card

print(my_card) # calls the "__str__" method on my_card
print(isinstance(my_card, CreditCard)) # allows you to check if an instance belongs to a specific class 


################ Inheritance ################
print("------------------ Inheritance ------------------")
class RewardCard(CreditCard): # RewardCard inherits from CreditCard
    def __init__(self, yearly_rate, miles_per_dollar):
        # because RewardCard inherits from CreditCard we are able to call the __init__ method of CreditCard here
        CreditCard.__init__(self, yearly_rate) # this initalises self.balance, self.yearly_rate, self.transactions 
        self.miles_per_dollar = miles_per_dollar
        self.miles = 0
        
    def get_miles(self):
        return self.miles
    
    def spend(self, amount):
        CreditCard.spend(self, amount)
        self.miles += amount * self.miles_per_dollar
        
    def __str__(self):
        return CreditCard.__str__(self)+"; reward "+str(self.get_miles())+" miles"

## what does RewardCard inherit from CreditCard? It inherits all of the methods from CreditCard
my_rewards_card = RewardCard(0.15, 2)
my_rewards_card.spend(1.99)
my_rewards_card.spend(3.99)
my_rewards_card.pay_bill(50) # is this legal? yes because it inherits the method from CreditCard
print(my_rewards_card)

## Challenge: Create your own data type using classes

################ Review ################
print("------------------ Review ------------------")
################  Tuples Review  ################
print("------------------ Tuples ------------------")
t = (1,2,3)
# how do I:
# - get the length of t
# - get the maximum element of t
# - get the element at the 1st index of t
# - get the last element of t 
# - get the first 2 elements of t
# - go through the elements of t
# - add 4 to t
# - delete the element at index 2


# Answers to tuples review: 
print("Answers for tuple review")
print("The length of t is", len(t))
print("The maximum element of t is", max(t))
print("The element at the 1st index of t is", t[1])
print("The last element of t is:", t[-1]) 
print("Another way to get the last element of t is:", t[len(t)-1]) # because python is 0 indexed, we need to subtract 1 from the length
print("The first two elements of t are:", t[:2]) # Note: the default start parameter is 0, so leaving it empty means the default value will be used
print("Here is our first approach of going through the elements of t")
# directly goes through the elements of t
for elem in t:
    print("Approach 1", elem)
print("Here is our second approach of going through the elements of t")
# goes through the indexes in t 
for i in range(len(t)):
    print("Approach 2", t[i])
print("A couple of trick questions here, tuples are immutable so we can't add or remove from them!")
print("Instead of adding to t, we can create a new tuple that has the elements of t and 4", t + (4,))
print("Instead of deleting the element at index 2 from t, we can create a new tuple using slicing", t[:2] + t[3:])


################  Lists Review  ################
print("------------------ Lists Review ------------------")
l = [1,2,3,4,5]
# how do I:
# - get the length of l
# - get the maximum element of l
# - get the element at the 1st index of l
# - get the last element of l
# - get the first 2 elements of l 
# - go through the elements of l
# - add 6 to the end of l
# - delete the element at index 2 of l


# Answers to list review: 
print("Answers for list review")
print("The length of l is", len(l))
print("The maximum element of l is", max(l))
print("The element at the 1st index of l is", l[1])
print("The last element of l is:", l[-1]) 
print("Another way to get the last element of l is:", l[len(l)-1]) # because python is 0 indexed, we need to subtract 1 from the length
print("The first two elements of l are:", l[:2]) # Note: the default start parameter is 0, so leaving it empty means the default value will be used
print("Here is our first approach of going through the elements of l")
# directly goes through the elements of l
# you should be careful to not mutate a list as you are iterating over it, this could cause funky behavior
for elem in l:
    print("Approach 1", elem)
print("Here is our second approach of going through the elements of l")
# goes through the indexes in l: generally safe because the range is only computed once in the beginning
for i in range(len(l)):
    print("Approach 2", l[i])

# Adding an element to a list
l.append(6) # note that is function mutates l, but doesn't return anything
print("L has an additional element not", l)

# deleting the second element of l
del l[2] # note that is function mutates l, but doesn't return anything
print("We have deleted the element at index 2", l)


################  Dictionaries Review  ################
print("------------------ Dictionaries Review ------------------")
d = {"day":"friday", "month": "february"}
# how do I:
# - get the length of d
# - get the maximum key in of d
# - get the maximum value in of d
# - update the "day" to be "monday"
# - add a new key: value pairing for "year" -> 2020
# - remove the key: value pairing "month": "february"
# - go through all of the key value pairings

print("The length of d is", len(d))
print("The maximum key in d is", max(d.keys()))
print("The maximum value in d is", max(d.values()))
# we can update the a key value pairing by doing
d["day"] = "monday"
print("updated the value associated with day", d)
# adding a new key value pairing
d["year"] = 2020
print("we've added a new key (year) and value (2019) to d", d)
# remove a key value pairing
del d["month"]
print("we've removed the key value pairing for month", d)
# go through all key value pairings in d
# approach 1: goes throug the keys of d 
for key in d.keys():
    print("Approach 1: key", key, ": value", d[key])
# approach 2: goes throug the key value pairings in d 
for pair in d.items():
    print("Approach 2: key", pair[0], ": value", pair[1])

################  Aliasing Review  ################
d = {"day":"friday", "month": "february"}
d_new = d
d["class"] = "6.0001 Recitation"
# what is d at this point?
# what is d_new at this point?

print("Answers for aliasing review")
print("d has been updated", d)
print("d_new has also been updated, because they both point to the same location in memory", d_new)
